/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <paths.h>
#include <limits.h>
#include <stdarg.h>
#include <errno.h>
#include <sysexits.h>
#include "faxrecv.h"
#include "ConfPars.h"
#include "atomar.h"
#include "osImport.h"

/*---------------------------------------------------------------------------*\
    from capiutils.h:
\*---------------------------------------------------------------------------*/

extern "C" char *capi_info2str(c_word reason);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define FAX_FILTER_BUFFER_SIZE      1000
#define CAPI_BUFFER_SIZE            CapiMaxDataSize

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

static void detachIO(void) {
    //endpwent();               // XXX some systems hold descriptors
    int fd = open (_PATH_DEVNULL, O_RDWR);
    dup2 (fd, STDIN_FILENO);
    dup2 (fd, STDOUT_FILENO);
    dup2 (fd, STDERR_FILENO);
    for (fd = _POSIX_OPEN_MAX - 1; (fd >= 0); fd--) {
        if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO) {
            close(fd);
        }
    }
}

void setRealIDs (void) {
    uid_t euid = geteuid();
    seteuid (0);
    setgid (getegid());
    setuid (euid);
}

tBool runCmd (char* cmd, bool changeIDs) {
    pid_t pid = fork();
    switch (pid) {
    case 0:
        if (changeIDs) {
            setRealIDs();
        }
        detachIO();
        execl ("/bin/sh", "sh", "-c", cmd, 0);
        sleep (1);
        _exit (127);
    default: {
        int status = 0;
        waitpid (pid, &status, 0);
        if (status == 0) {
            return vTrue;
        }
        break; }
    case -1:
        break;
    }
    return vFalse;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

tUInt FaxPropEncode (tUInt vRes, tUInt BaudRate) {
    //params.encode:          st|bf|ec|df|ln| wd| br|vr
    //values for capi:       000  0  0 00 10 000 ??? 0/1
    tUInt br = 0;
    if (BaudRate > 12000) {
        br = 5;
    } else if (BaudRate > 9600) {
        br = 4;
    } else if (BaudRate > 7200) {
        br = 3;
    } else if (BaudRate > 4800) {
        br = 2;
    } else if (BaudRate > 2400) {
        br = 1;
    }
    return 0x100 + (br << 1) + (vRes ? 1 : 0);
}

/*===========================================================================*\
\*===========================================================================*/

CFaxReceive::CFaxReceive (CFaxReceiveDevice *FaxRecvDevice, FaxFormatType faxtype) {
    dhead ("CFaxReceive - Constructor", DEBUG_CFaxReceive);
    FaxDevice   = FaxRecvDevice;
    format      = faxtype;
    RecvNames   = 0;
    RecvFile    = -1;
    RecvString  = 0;
    ONumber     = 0;
    MyNumber    = 0;
    faxFileNr   = 0;
    jobNr       = 0;
    NumPages    = 0;
    VResolution = 0;
    BaudRate    = 0;
    recvState   = 0;
    quitRecv    = 0;
    StartTime   = 0;
    hJobFile    = 0;
}

/*===========================================================================*\
\*===========================================================================*/

CFaxReceive::~CFaxReceive (void) {
    dhead ("CFaxReceive - Destructor", DEBUG_CFaxReceive);
    dwarning (RecvFile == -1);
    if (RecvFile >= 0) {
        close (RecvFile);
        RecvFile = -1;
    }

    if (hJobFile > 0) {
        fclose (hJobFile);
        hJobFile = 0;
    }
    Conv_DuplicateString (&ONumber, 0);
    Conv_DuplicateString (&MyNumber, 0);
    Conv_DuplicateString (&RecvString, 0);
    DeleteRecvNames();
    SffFilter.Close();
}


/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceive::StartReceive (void) {
    dhead ("CFaxReceive::StartReceive", DEBUG_CFaxReceive);
    if ((format != FaxFormat_SFF) && (SffFilter.Open (1) == vFalse)) {
        return vFalse;
    }
    recvState = 0;

    cu_B3Configuration B3Config;
    memset (&B3Config, '\0', sizeof (B3Config));
    B3Config.Options    = 1;
    B3Config.Format     = 0;
    B3Config.StationID  = FaxDevice->FaxNumber.GetPointer();
    B3Config.HeadLine   = 0; //Identifier.GetPointer();

    SetB1Parameters (b1prot_G3Fax);
    SetB2Parameters (b2prot_G3Fax);
    SetB3Parameters (b3prot_G3Fax_extended, &B3Config);
    SetNCPI (b3prot_G3Fax, 0);

    CheckListen();
    return vTrue;
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::StopReceive (void) {
    dhead ("CFaxReceive::StopReceive", DEBUG_CFaxReceive);
    quitRecv = 1;
    ShareMSNList (0);
    Disconnect();
    CheckListen();
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::IncomingCall (tUInt /*Controller*/, c_cipvalue CIPValue, tUChar *pOppositeNumber, tUChar *pMyCalledNumber) {
    dhead ("CFaxReceive::IncomingCall", DEBUG_CFaxReceive);
    dparams ("%x,%s,%s", CIPValue, pOppositeNumber, pMyCalledNumber);
    faxFileNr   = 0;
    jobNr       = 0;
    NumPages    = 0;
    VResolution = 0;
    BaudRate    = 0;
    StartTime   = 0;
    Conv_DuplicateString (&ONumber, pOppositeNumber);
    Conv_DuplicateString (&MyNumber, pMyCalledNumber);
    Conv_DuplicateString (&RecvString, 0);

    char *fromtext = " from ";
    char *totext   = " to ";
    char *ciptext  = "fax";
    switch (CIPValue) {
    case cipvalue_Speech:
        ciptext = "speech";
        break;
    case cipvalue_3_1kHzAudio:
        ciptext = "analog";
        break;
    default:
        // nothing to do
        break;
    }
    if (!pOppositeNumber || !*pOppositeNumber) {
        pOppositeNumber = (tUChar *)"";
        fromtext        = "";
    }
    if (!pMyCalledNumber || !*pMyCalledNumber) {
        pMyCalledNumber = (tUChar *)"";
        totext          = "";
    }
    if (format == FaxFormat_Hylafax) {
        if (GetJobFile() == vFalse) {
            WriteLog (LOG_ERR, "Cant create log file for incomming call!\n");
            // TODO: better disconnect the call?
        }
        WriteLog (LOG_INFO, "SESSION BEGIN %08u %s\n", jobNr, FaxDevice->FaxNumber.GetPointer());
    }
    WriteLog (LOG_INFO, "Incoming %s call%s%s%s%s.\n", ciptext, fromtext, pOppositeNumber, totext, pMyCalledNumber);
    FaxDevice->WriteStatus ("Answering the phone");
    AnswerCall (vTrue);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::DoConnecting (void) {
    dhead ("CFaxReceive::DoConnecting", DEBUG_CFaxReceive);
    // Nothing to do!
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::ConnectProceeding (void) {
    dhead ("CFaxReceive::ConnectProceeding", DEBUG_CFaxReceive);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::IsConnected (cu_NCPI *pNCPI) {
    dhead ("CFaxReceive::IsConnected", DEBUG_CFaxReceive);
    char internaltext[MAX_STRING_SIZE] = "";
    if (pNCPI) {
        if (  (pNCPI->Format != 0)          // no SFF format
           || (pNCPI->Options & 0x2)) {     // fax polling not implemented yet
            dwarning (0);
            Disconnect();
            return;
        }
        VResolution = pNCPI->Options & 1;
        BaudRate    = pNCPI->Rate;
        Conv_DuplicateString (&RecvString, pNCPI->ReceiveID);
        dprint ("recvID=%s(%u) ", RecvString, strlen ((char *)RecvString));
        if (RecvString && *RecvString) {
            sprintf (internaltext, " to Receive ID '%s'", RecvString);
        }
        if (BaudRate) {
            sprintf (internaltext + strlen (internaltext), " with %u bit/s", BaudRate);
        }
    }
    StartTime = time (0);
    if (  (GetRecvFile() < 0)
       || (  (  (format == FaxFormat_Hylafax)
             || (format == FaxFormat_TIFF))
          && (TifFilter.Open (RecvFile, RecvNames->string) == vFalse))) {
        WriteLog (LOG_ERR, "Cant create file for the fax. Disconnecting now!\n");
        DeleteRecvNames();
        Disconnect();
    } else {
        NotifyFIFO ('I');
        NotifyFIFO ('B');
        FaxDevice->WriteStatus ("Receiving facsimile");
        NotifyFIFO ('S');
        recvState = 1;
        WriteLog (LOG_INFO, "Connection established%s.\n", internaltext);
        WriteLog (LOG_INFO, "Write fax to file %s.\n", RecvNames->string);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::PutDataComplete (void *, c_info) {
    dhead ("CFaxReceive::PutDataComplete", DEBUG_CFaxReceive);
    dwarning (0); // Should never been called!
    // Nothing to do!
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::GetData (tUByte *Data, tUShort DataLength, tUShort DataHandle) {
    dhead ("CFaxReceive::GetData", Dbg_Level_Infos); //DEBUG_CFaxReceive);
    dassert (Data != 0);
    dwarning (DataLength > 0);
    if (format == FaxFormat_SFF) {
        write (RecvFile, Data, DataLength);
        GetDataComplete (DataHandle);
        return;
    }
    switch (recvState) {
    default:
        GetDataComplete (DataHandle);
        return;
    case 1: {
        CSFilter_Sff2Huf::ctPageInfo PageInfo;
        for (; DataLength > 0; DataLength--, Data++) {
            // TODO: Fehler von PutChar abfangen
            SffFilter.PutChar (*Data);
            if ((recvState == 1) && (SffFilter.GetCurPageInfo (&PageInfo) == vTrue)) {
                tUChar hostname[MAX_STRING_SIZE];
                gethostname ((char *)hostname, MAX_STRING_SIZE);
                VResolution = PageInfo.ResolutionVertical;
                TifFilter.PutPageHeader ((VResolution) ? 196 : 98, 203,
                                         PageInfo.PageLength, PageInfo.LineLength,
                                         FaxPropEncode (VResolution, BaudRate),
                                         (RecvString && *RecvString) ? RecvString : ONumber,
                                         (tUChar *)"AVM GmbH", (tUChar *)NAME_OF_MODEL, hostname,
                                         (tUChar *)NAME_OF_SOFTWARE);
                recvState = 2;
            }
        }
        break; }
    case 2:
        for (; DataLength > 0; DataLength--, Data++) {
            // TODO: Fehler von PutChar abfangen
            SffFilter.PutChar (*Data);
        }
        break;
    }
    GetDataComplete (DataHandle);

    CSFilter_Sff2Huf::ctInfo info;
    tUInt len;
    tUChar buf[1000];
    do {
        len = SffFilter.GetToPageEnd (buf, 1000, &info);
        if (format == FaxFormat_G3) {
            write (RecvFile, buf, len);
            switch (info) {
            case CSFilter_Sff2Huf::i_NewPage:
                NumPages++;
                if (GetNextRecvFile() < 0) {
                    WriteLog (LOG_ERR, "Cant create new file for next fax page!\n");
                    // TODO: How to handle this? Set quitRecv?
                    recvState  = 3;
                    len        = 0;  // to stop the loop
                    Disconnect();
                }
                break;
            case CSFilter_Sff2Huf::i_LastPage:
                NumPages++;
                close (RecvFile);
                RecvFile  = -1;
                recvState = 3;
                len       = 0;      // to stop the loop
                break;
            default:
                // nothing to do
                break;
            }
        } else {
            if (len) {
                TifFilter.Put (buf, len);
            }
            switch (info) {
            case CSFilter_Sff2Huf::i_NewPage: {
                CSFilter_Sff2Huf::ctPageInfo PageInfo;
                SffFilter.GetLastPageInfo (&PageInfo);
                TifFilter.PutPageEnd (vFalse, PageInfo.PageLength, time (0) - StartTime);
                SffFilter.GetCurPageInfo (&PageInfo);
                NumPages++;
                NotifyFIFO ('P');
                VResolution = PageInfo.ResolutionVertical;
                TifFilter.PutPageHeader ((VResolution) ? 196 : 98, 203,
                                         PageInfo.PageLength, PageInfo.LineLength,
                                         0, 0, 0, 0, 0, 0);
                len = 1; // to continue the loop
                break;
            }
            case CSFilter_Sff2Huf::i_LastPage: {
                NumPages++;
                NotifyFIFO ('P');
                CSFilter_Sff2Huf::ctPageInfo PageInfo;
                SffFilter.GetLastPageInfo (&PageInfo);
                TifFilter.PutPageEnd (vTrue, PageInfo.PageLength, time (0) - StartTime);
                recvState = 3;
                len = 0; // to stop the loop
                break;
            }
            default:
                // nothing to do
                break;
            }
        }
    } while (len);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::DoDisconnecting (c_info, cu_NCPI *pNCPI) {
    dhead ("CFaxReceive::DoDisconnecting", DEBUG_CFaxReceive);
    if (pNCPI) {
        NumPages    = pNCPI->Pages;
        VResolution = pNCPI->Options & 1;
        BaudRate    = pNCPI->Rate;
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::IsDisconnected (c_info Reason) {
    dhead ("CFaxReceive::IsDisconnected", DEBUG_CFaxReceive);
    dparams ("%x", Reason);
    if (RecvFile >= 0) {
        close (RecvFile);
        RecvFile = -1;
    }

    if (CapiInfoIsRegError (Reason)) {
        // Get a Register-Errror as Reason => CAPI not installed, terminate Programm
        dprint ("\nCAPI not installed! ");
        WriteLog (LOG_ERR, "CAPI not installed, started or have no access rights on it!\n");
        FaxDevice->WriteStatus ("CAPI not installed, started or have no access rights on it!!");
        NotifyFIFO ('W');
        quitRecv   = 1;
    } else {
        switch (format) {
        case FaxFormat_Hylafax:
        case FaxFormat_TIFF:
            TifFilter.Close();
            // no break!
        case FaxFormat_G3:
            SffFilter.Close();
            if (SffFilter.Open (1) == vFalse) {
                //ListenRemove (0);
                WriteLog (LOG_ERR, "Allocating memory error! Stop Listening!\n");
                FaxDevice->WriteStatus ("Allocating memory error! Stop Listening!");
                NotifyFIFO ('W');
                quitRecv   = 1;
            } else {
                NotifyFIFO ('E');
            }
            break;
        default:
            // nothing to do
            break;
        }
        if (RecvNames) {
            if (FaxDevice->MainClass->FaxRcvdCmd.IsEmpty() == vFalse) {
                CDynamicString executeCommand;
                executeCommand.Set (&FaxDevice->MainClass->FaxRcvdCmd);
                tUChar numberStr[20];

                if (format == FaxFormat_Hylafax) {
                    char *StateText = "";
                    switch (recvState) {
                    default:
                    case 0:
                        StateText = "Can't build up connection for incoming call.";
                        break;
                    case 1:
                        StateText = "Connection was established but no fax data was received.";
                        break;
                    case 2:
                        StateText = "Connection droped unexpected during fax receiving.";
                        break;
                    case 3:
                        break;
                    }

                    // Hylafax: <qfile> <ModemDeviceID> <CommID> <Reason>
                    sprintf ((char *)numberStr, "\" \"%08u\" \"", jobNr);
                    executeCommand.Append (" \"");
                    executeCommand.Append (RecvNames->string);
                    executeCommand.Append ("\" \"");
                    executeCommand.Append (&FaxDevice->DeviceName);
                    executeCommand.Append (numberStr);
                    executeCommand.Append (StateText);
                    executeCommand.Append ("\"");
                } else  {
                    // mgetty:  <RecvStatus> <Hangup Code> "<sender id>" "<receiver id>" <nr of pages> <file(s)>
                    tUInt recvStatus = 0;
                    switch (recvState) {
                    default:
                    case 0:
                        recvStatus = 6;
                        break;
                    case 1:
                        recvStatus = 7;
                        break;
                    case 2:
                        recvStatus = 12;
                        break;
                    case 3:
                        break;
                    }
                    sprintf ((char *)numberStr, " %u ", recvStatus);
                    executeCommand.Append (numberStr);
                    sprintf ((char *)numberStr, "0x%X \"", Reason);
                    executeCommand.Append (numberStr);
                    if (RecvString && *RecvString) {
                        executeCommand.Append (RecvString);
                    } else if (ONumber && *ONumber) {
                        executeCommand.Append (ONumber);
                    }
                    executeCommand.Append ("\" \"");
                    if (MyNumber) {
                        executeCommand.Append (MyNumber);
                    }
                    sprintf ((char *)numberStr, "\" %u", NumPages);
                    executeCommand.Append (numberStr);
                    for (NameList *pLauf = RecvNames; (pLauf != 0); pLauf = pLauf->next) {
                        executeCommand.Append (" \"");
                        executeCommand.Append (pLauf->string);
                        executeCommand.Append ("\"");
                    }
                }

                if (executeCommand.DetectAnError() == vTrue) {
                    WriteLog (LOG_INFO, "Fax received and stored under %s.\n", RecvNames->string);
                    WriteLog (LOG_ERR, "Allocating memory error! Cant notify new received fax!\n");
                    FaxDevice->WriteStatus ("Allocating memory error! Cant notify new received fax!");
                    NotifyFIFO ('W');
                    quitRecv = 1;
                } else {
                    WriteLog (LOG_INFO, "Fax received and calling '%s'.\n", executeCommand.GetPointer());
                    runCmd ((char *)executeCommand.GetPointer(), vTrue);
                }
            } else {
                WriteLog (LOG_INFO, "Fax received and stored under %s.\n", RecvNames->string);
            }
        }
        WriteLog (LOG_INFO, "Connection is droped with reason 0x%X (%s).\n", Reason, capi_info2str (Reason));
        if (hJobFile > 0) {
            WriteLog (LOG_INFO, "SESSION END\n");
            fclose (hJobFile);
            hJobFile = 0;
        }
    }
    recvState   = 0;
    faxFileNr   = 0;
    jobNr       = 0;
    NumPages    = 0;
    VResolution = 0;
    StartTime   = 0;
    Conv_DuplicateString (&ONumber, 0);
    Conv_DuplicateString (&MyNumber, 0);
    Conv_DuplicateString (&RecvString, 0);
    FaxDevice->WriteStatus ("Running and idle");
    NotifyFIFO ('I');
    NotifyFIFO ('R');
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::IsReleased (c_info /*Info*/) {
    dhead ("CFaxReceive::IsReleased", DEBUG_CFaxReceive);
    //dassert (CapiInfoIsSuccess (Info));
    if (quitRecv == vFalse) {
        WriteLog (LOG_ERR, "Can't register at CAPI! Is CAPI loaded and accessible from this program?\n");
    } else {
        quitRecv = vFalse;
    }
    CFaxReceiveDevice *fd = FaxDevice;
    delete this;
    fd->ReceiveStoped();
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceive::NotifyFIFO (char Type) {
    dhead ("CFaxReceive::NotifyFIFO", DEBUG_CFaxReceive);
    if (format != FaxFormat_Hylafax) {
        return vTrue;
    }
    char   FaxInfo[4 * MAX_STRING_SIZE];
    tBool  recv = vFalse;
    char  *text = 0;
    switch (Type) {
    // Modem part
    case 'I':
        if (jobNr > 0) {
            sprintf (FaxInfo, "%08u", jobNr);
            text = FaxInfo;
        }
        break;
    case 'C':
        // TODO? MSN eintragen?
        text = (char *)ONumber;
        break;

    // Receive part
    case 'S':
    case 'P':
    case 'D':
        sprintf (FaxInfo, "%lx,%x,%x,%s,%08u,\"%s\",\"%s\",\"%s\"", time (0) - StartTime, NumPages,
                 FaxPropEncode (VResolution, BaudRate), RecvNames->string, jobNr,
                 RecvString, ONumber, "");
        text = FaxInfo;
        recv = vTrue;
        break;
    default:
        return FaxDevice->NotifyFIFO (Type);
    }
    return FaxDevice->NotifyFIFO (recv, Type, text);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::WriteLog (tSInt priority, char *text, ...) {
    dhead ("CFaxReceive::WriteLog", DEBUG_CFaxReceive);
    char    timebuf[MAX_STRING_SIZE];
    timeval tv;
    va_list params;
    va_start (params, text);

    gettimeofday (&tv, 0);
    strftime (timebuf, MAX_STRING_SIZE, "%h %d %T", localtime((time_t*) &tv.tv_sec));
    sprintf (timebuf + strlen(timebuf), ".%02lu: [%5d]: ", tv.tv_usec / 10000, getpid());

    if (hJobFile) {
        fprintf (hJobFile, timebuf);
        vfprintf (hJobFile, text, params);
    }

    FaxDevice->vWriteLog (priority , text, params);
    va_end (params);
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceive::GetJobFile (void) {
    dhead ("CFaxReceive::GetJobFile", DEBUG_CFaxReceive);
    dassert (format == FaxFormat_Hylafax);
    jobNr = 0;
    if (hJobFile != 0) {
        fclose (hJobFile);
        hJobFile = 0;
    }
    char  astring[MAX_STRING_SIZE];
    tSInt len;
    tUInt tmpJobFileNr = 1;
    tSInt tmpFileD     = -1;
    tSInt SeqFile      = open (HYLAFAX_LOG_DIR HYLAFAX_SEQF_FILE, O_CREAT | O_RDWR, 0644);
    if (SeqFile >= 0) {
        flock (SeqFile, LOCK_EX);
        len = read (SeqFile, astring, MAX_STRING_SIZE);
        astring[(len <= 0) ? 0 : len] = '\0';
        tmpJobFileNr = atoi (astring) + 1;
    }
    for (tUInt trys = MAX_TRY_COUNT; trys > 0; trys--, tmpJobFileNr++) {
        if (tmpJobFileNr < 1 || tmpJobFileNr > MAX_SEQ_NUMBER) {
            tmpJobFileNr = 1;
        }
        sprintf (astring, HYLAFAX_LOG_DIR "c%08u", tmpJobFileNr);
        tmpFileD = open (astring, O_RDWR | O_CREAT | O_EXCL, FaxDevice->MainClass->LogFileMode);
        dprint ("try(%s,%u,%d) ", astring, strlen (astring), tmpFileD);
        if (tmpFileD >= 0) {
            flock (tmpFileD, LOCK_EX | LOCK_NB);
            hJobFile = fdopen (tmpFileD, "w");
            dprint ("fdopen(%x,%d) ", hJobFile, tmpFileD);
            if (hJobFile != 0) {
                jobNr = tmpJobFileNr;
                if (SeqFile >= 0) {
                    sprintf (astring , "%u", tmpJobFileNr);
                    len = strlen (astring);
                    lseek (SeqFile, 0, SEEK_SET);
                    write (SeqFile, astring, len);
                    ftruncate (SeqFile, len);
                }
                break;
            }
            close (tmpFileD);
            tmpFileD = -1;
            continue;
        }
        if (errno != EEXIST) {
            break;
        }
    }
    if (SeqFile >= 0) {
        close (SeqFile);
    }
    return (hJobFile != 0);
}

/*===========================================================================*\
\*===========================================================================*/

tSInt CFaxReceive::GetRecvFile (void) {
    dhead ("CFaxReceive::GetRecvFile", DEBUG_CFaxReceive);
    tUInt tmpfaxFileNr                   =  1;
    tSInt tmpFileD                       = -1;
    tSInt SeqFile                        = -1;
    tSInt len;
    char  NewNameFormat[MAX_STRING_SIZE] = "";
    char  NewName[MAX_STRING_SIZE];
    char  fline[MAX_STRING_SIZE];
    DeleteRecvNames();
    if (RecvFile >= 0) {
        close (RecvFile);
        RecvFile = -1;
    }
    if (format == FaxFormat_Hylafax) {
        strcpy (NewNameFormat, HYLAFAX_RECV_DIR);
    } else if (FaxDevice->MainClass->SpoolDir.IsEmpty() == vFalse) {
        strcpy (NewNameFormat, (char *)FaxDevice->MainClass->SpoolDir.GetPointer());
        if (NewName[strlen (NewNameFormat) - 1] != '/') {
            strcat (NewNameFormat, "/");
        }
    }
    strcpy (fline, NewNameFormat);
    strcat (fline, HYLAFAX_SEQF_FILE);
    SeqFile = open (fline, O_CREAT | O_RDWR, 0644);
    if (SeqFile >= 0) {
        flock (SeqFile, LOCK_EX);
        len = read (SeqFile, fline, MAX_STRING_SIZE);
        fline[(len < 0) ? 0 : len] = '\0';
        tmpfaxFileNr = atoi (fline) + 1;
    }
    if (format == FaxFormat_Hylafax) {
        strcat (NewNameFormat, "fax%05u" FILE_EXTENSION_TIFF);
    } else {
        sprintf (NewNameFormat + strlen (NewNameFormat), "cf%c%%07u", (VResolution) ? 'f' : 'n');
        if (RecvString && *RecvString) {
            strcat (NewNameFormat, "-");
            strcat (NewNameFormat, (char *)RecvString);
        } else if (ONumber && *ONumber) {
            strcat (NewNameFormat, "-");
            strcat (NewNameFormat, (char *)ONumber);
        }
        switch (format) {
        case FaxFormat_TIFF:
            strcat (NewNameFormat, FILE_EXTENSION_TIFF);
            break;
        case FaxFormat_SFF:
            strcat (NewNameFormat, FILE_EXTENSION_SFF);
            break;
        case FaxFormat_G3:
            strcat (NewNameFormat, ".01" FILE_EXTENSION_G3);
            break;
        default:
            // nothing to do
            break;
        }
    }
    for (tUInt trys = MAX_TRY_COUNT; trys > 0; trys--, tmpfaxFileNr++) {
        if (tmpfaxFileNr < 1 || tmpfaxFileNr > MAX_SEQ_NUMBER) {
            tmpfaxFileNr = 1;
        }
        sprintf (NewName, NewNameFormat, tmpfaxFileNr);
        tmpFileD = open (NewName, O_RDWR | O_CREAT | O_EXCL, FaxDevice->RecvFileMode);
        dprint ("try(%s,%u,%d) ", NewName, strlen (NewName), tmpFileD);
        if (tmpFileD >= 0) {
            flock (tmpFileD, LOCK_EX | LOCK_NB);
            RecvNames = (NameList *)new tUByte [strlen (NewName) + sizeof (void *) + sizeof (tUChar)];
            if (!RecvNames) {
                close (tmpFileD);
                break;
            }
            strcpy ((char *)RecvNames->string, NewName);
            RecvNames->next = 0;
            faxFileNr = tmpfaxFileNr;
            RecvFile  = tmpFileD;
            if (SeqFile >= 0) {
                sprintf (fline , "%u", faxFileNr);
                len = strlen (fline);
                lseek (SeqFile, 0, SEEK_SET);
                write (SeqFile, fline, len);
                ftruncate (SeqFile, len);
            }
            break;
        }
        if (errno != EEXIST) {
            break;
        }
    }
    if (SeqFile >= 0) {
        close (SeqFile);
    }
    return RecvFile;
}

/*===========================================================================*\
\*===========================================================================*/

tSInt CFaxReceive::GetNextRecvFile (void) {
    dhead ("CFaxReceive::GetNextRecvFile", DEBUG_CFaxReceive);
    dassert (format == FaxFormat_G3);
    dassert (faxFileNr > 0);
    dassert (RecvNames != 0);
    tSInt tmpFileD                       = -1;
    char  NewNameFormat[MAX_STRING_SIZE] = "";
    char  NewName[MAX_STRING_SIZE];

    if (RecvFile >= 0) {
        close (RecvFile);
        RecvFile = -1;
    }
    if (FaxDevice->MainClass->SpoolDir.IsEmpty() == vFalse) {
        strcpy (NewNameFormat, (char *)FaxDevice->MainClass->SpoolDir.GetPointer());
        if (NewName[strlen (NewNameFormat) - 1] != '/') {
            strcat (NewNameFormat, "/");
        }
    }
    //sprintf (NewNameFormat + strlen (NewNameFormat), "cf%c%07u", (VResolution) ? 'f' : 'n', faxFileNr);
    sprintf (NewNameFormat + strlen (NewNameFormat), "cf%07u", faxFileNr);
    if (RecvString && *RecvString) {
        strcat (NewNameFormat, "-");
        strcat (NewNameFormat, (char *)RecvString);
    } else if (ONumber && *ONumber) {
        strcat (NewNameFormat, "-");
        strcat (NewNameFormat, (char *)ONumber);
    }
    strcat (NewNameFormat, ".%02u");
    strcat (NewNameFormat, FILE_EXTENSION_G3);
    // TODO?: can we try to open only the first specified file?
    for (tUInt nr = NumPages + 1; nr < NumPages + MAX_TRY_COUNT; nr++) {
        sprintf (NewName, NewNameFormat, nr);
        tmpFileD = open (NewName, O_RDWR | O_CREAT | O_EXCL, FaxDevice->RecvFileMode);
        if (tmpFileD >= 0) {
            flock (tmpFileD, LOCK_EX | LOCK_NB);
            NameList *NewLE = (NameList *)new tUByte [strlen (NewName)+sizeof (void *)+sizeof (tUChar)];
            if (!NewLE) {
                close (tmpFileD);
                break;
            }
            strcpy ((char *)NewLE->string, NewName);
            NewLE->next     = 0;
            NameList *pLauf = RecvNames;
            while (pLauf->next != 0) {
                pLauf = pLauf->next;
            }
            pLauf->next = NewLE;
            RecvFile    = tmpFileD;
            WriteLog (LOG_INFO, "File for next fax page is %s.\n", NewName);
            break;
        }
        if (errno != EEXIST) {
            break;
        }
    }
    return RecvFile;
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceive::DeleteRecvNames (void) {
    dhead ("CFaxReceive::DeleteRecvNames", DEBUG_CFaxReceive);
    NameList *delName;
    while (RecvNames) {
        delName   = RecvNames;
        RecvNames = RecvNames->next;
        delete [] delName;
    }
}

/*===========================================================================*\
\*===========================================================================*/


