/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "recvdev.h"
#include "faxrecv.h"
#include "defaults.h"
#include "ConfAssi.h"
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>

/*===========================================================================*\
\*===========================================================================*/

void closeFIFO (int devinfo, const char * fifoName = 0) {
    if (devinfo != -1) {
        close (devinfo);
    }
    if (fifoName != 0) {
        unlink (fifoName);
    }
}

int openFIFO(const char* fifoName, int mode) {
    closeFIFO (-1, fifoName);
    if (  (access (fifoName, F_OK) == 0)
       || (mkfifo (fifoName, mode & 0777) < 0)) {
        return -1;
    }
    int fd = open (fifoName, O_RDWR | O_NDELAY, mode & 0777);
    return (fd);
}

/*===========================================================================*\
\*===========================================================================*/

CFaxReceiveDevice::CFaxReceiveDevice (CFaxReceiveMain *MainFaxClass, FaxFormatType faxtype) {
    dhead ("CFaxReceiveDevice-Constructor", DEBUG_CFaxReceiveDevice);
    DeviceName.Set (DEFAULT_DEVICE_NAME);
    MainClass      = MainFaxClass;
    format         = faxtype;
    RecvFileMode   = DEFAULT_FILE_MODE;
    CountThreads   = 2;
    hFIFOfile      = -1;
    Starting       = vFalse;
}

/*===========================================================================*\
\*===========================================================================*/

CFaxReceiveDevice::~CFaxReceiveDevice (void) {
    dhead ("CFaxReceiveDevice-Destructor", DEBUG_CFaxReceiveDevice);
    if (hFIFOfile >= 0) {
        closeFIFO (hFIFOfile);
        hFIFOfile = -1;
    }
    if (hStatusFile > 0) {
        fclose (hStatusFile);
        hStatusFile = 0;
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::Config (CConfigParserSection *section) {
    dhead ("CFaxReceiveDevice::Config", DEBUG_CFaxReceiveDevice);
    dassert (section != 0);
    ConfigAssignment_Item Assignments[] = {
        { "HylafaxDeviceName",  ConfigAssignment_String, &DeviceName },
        { "FAXNumber",          ConfigAssignment_String, &FaxNumber },
        { "RecvFileMode",       ConfigAssignment_UInt,   &RecvFileMode },
        { "CountThreads",       ConfigAssignment_UInt,   &CountThreads },
        { 0,                    ConfigAssignment_UInt,   0 }
    };
    ConfigValuesAssignment (&section->ValueList, &Assignments[0]);

    tUInt         tmpController;
    tUInt         tmpUseDDI;
    tUInt         tmpDDILen;
    tUInt         tmpDDIOffset;
    tUInt         tmpSpeech;
    tUInt         tmpCIPMask;
    CPointerList  MSNs;

    ConfigAssignment_Item CntrlAssignments[] = {
        { "Controller",   ConfigAssignment_UInt,       &tmpController },
        { "UseDDI",       ConfigAssignment_UInt,       &tmpUseDDI },
        { "DDILength",    ConfigAssignment_UInt,       &tmpDDILen },
        { "DDIOffset",    ConfigAssignment_UInt,       &tmpDDIOffset },
        { "AcceptSpeech", ConfigAssignment_UInt,       &tmpSpeech },
        { "IncomingMSNs", ConfigAssignment_StringList, &MSNs },
        { 0,              ConfigAssignment_UInt,       0 }
    };

    for (CConfigParserSection *laufSec = (CConfigParserSection *)section->SectionList.GetFirst();
         laufSec; laufSec = (CConfigParserSection *)laufSec->GetNext()) {

        tmpController = 1;
        tmpUseDDI     = 0;
        tmpDDILen     = 0;
        tmpDDIOffset  = 0;
        tmpSpeech     = 0;
        tmpCIPMask    = cipmask_3_1kHzAudio | cipmask_G3Fax;
        MSNs.RemoveAndDeleteAll();
        ConfigValuesAssignment (&laufSec->ValueList, &CntrlAssignments[0]);
        if ((tmpController != 0) && (CIPMSNList.ExistController (tmpController) == vFalse)) {
            if (tmpSpeech) {
                tmpCIPMask |= cipmask_Speech;
            }
            if (tmpUseDDI != 0) {
                CIPMSNList.Set (tmpController, 0, tmpCIPMask, tmpDDIOffset, tmpDDILen);
            } else {
                CIPMSNList.Set (tmpController, 0, tmpCIPMask);
                if (MSNs.IsEmpty() == vFalse) {
                    CConfigParserString *MSNlauf = (CConfigParserString *)MSNs.GetFirst();
                    dassert (MSNlauf != 0);
                    dassert (MSNlauf->GetPointer() != 0);
                    CIPMSNList.SetMSN (tmpController, MSNlauf->GetPointer());
                    while ((MSNlauf = (CConfigParserString *)MSNlauf->GetNext()) != 0) {
                        dassert (MSNlauf->GetPointer() != 0);
                        CIPMSNList.AddMSN (tmpController, MSNlauf->GetPointer());
                    }
                }
            }
        }
    }
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceiveDevice::StartReceive (void) {
    dhead ("CFaxReceiveDevice::StartReceive", DEBUG_CFaxReceiveDevice);
    Starting = vTrue;
    if (format == FaxFormat_Hylafax) {
        char StatusFileName[MAX_STRING_SIZE];
        sprintf (StatusFileName, HYLAFAX_STATUS_DIR "%s", DeviceName.GetPointer());
        hStatusFile = fopen (StatusFileName, "w");
        if (hStatusFile) {
            fchmod (fileno(hStatusFile), MainClass->LogFileMode);
            WriteStatus ("Running and idle");
        }
    }

    for (tUInt i = CountThreads; i > 0; i--) {
        CFaxReceive *pFR = new CFaxReceive (this, format);
        if (pFR) {
            FaxThreads.AddLast (pFR);
            pFR->ShareMSNList (&CIPMSNList);
            pFR->StartReceive();
            if (FaxThreads.IsEmpty() == vTrue) {
                Starting = vFalse;
                return vFalse;
            }
        }
    }
    Starting = vFalse;
    return vTrue;
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::StopReceive (void) {
    dhead ("CFaxReceiveDevice::StopReceive", DEBUG_CFaxReceiveDevice);
    for (CFaxReceive *lauf = (CFaxReceive *)FaxThreads.GetFirst(); lauf != 0; lauf = (CFaxReceive *)lauf->GetNext()) {
        lauf->StopReceive();
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::ReceiveStoped (void) {
    dhead ("CFaxReceiveDevice::ReceiveStoped", DEBUG_CFaxReceiveDevice);
    WriteStatus ("Terminated");
    NotifyFIFO ('Q');

    if ((Starting == vFalse) && FaxThreads.IsEmpty()) {
        delete this;
    }
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceiveDevice::RecreateFIFO (void) {
    dhead ("CFaxReceiveDevice::RecreateFIFO", DEBUG_CFaxReceiveDevice);
    if (hFIFOfile >= 0) {
        closeFIFO (hFIFOfile);
        hFIFOfile = -1;
    }
    char FifoName[MAX_STRING_SIZE];
    sprintf (FifoName, READ_FIFO_PREFIX "%s", DeviceName.GetPointer());
    hFIFOfile = openFIFO (FifoName, 0666);
    if (hFIFOfile < 0) {
        return vFalse;
    }
    NotifyFIFO ('R');
    NotifyFIFO ('N');
    NotifyFIFO ('R');
    return vTrue;
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::FIFOchanged (void) {
    dhead ("CFaxReceiveDevice::FIFOchanged", DEBUG_CFaxReceiveDevice);
    dprint ("\n");
    char  ReadCmd[MAX_STRING_SIZE] = "";
    tUInt len                      = read (hFIFOfile, ReadCmd, MAX_STRING_SIZE);
    dprint ("select-read(%s,%x)\n", ReadCmd, len);
    if (len && (len + 1)) {
        switch (*ReadCmd) {
        //case 'A':
        //    NotifyFIFO ('I');
        //    AnswerCall (vTrue);
        //    break;
        //case 'C':
        //    SetConfigItem (ReadCmd + 1, ConfigParamsCount, ConfigParams);
        //    break;
        case 'H':
            NotifyFIFO ('N');
            NotifyFIFO ('R');
            break;
        case 'S':
            // TODO: Not implemented yet!
            break;
        case 'Q':
            StopReceive();
            break;
        //case 'Z':
        //    Disconnect();
        //    break;
        }
    }
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceiveDevice::NotifyFIFO (tBool Receive, char Type, char *Data) {
    dhead ("CFaxReceiveDevice::NotifyFIFO", DEBUG_CFaxReceive);
    tBool fret = vFalse;
    if (format != FaxFormat_Hylafax) {
        fret = vTrue;
    } else {
        int faxqfifo = open (WRITE_FIFO_FILE, O_WRONLY | O_NDELAY);
        if (faxqfifo >= 0) {
            char msg[MAX_STRING_SIZE];
            sprintf (msg, "%c%s:%c", (Receive == vFalse) ? '+' : '@', DeviceName.GetPointer(), Type);
            if (Data && *Data) {
                strcat (msg, Data);
            }
            tUInt len = strlen (msg) + 1;
            if ((tSInt)len <= write (faxqfifo, msg, len)) {
                fret = vTrue;
            }
            close (faxqfifo);
        }
    }
    RETURN ('x', fret);
}

/*===========================================================================*\
\*===========================================================================*/

tBool CFaxReceiveDevice::NotifyFIFO (char Type) {
    dhead ("CFaxReceiveDevice::NotifyFIFO", DEBUG_CFaxReceive);
    if (format != FaxFormat_Hylafax) {
        return vTrue;
    }
    tBool  recv = vFalse;
    char  *text = 0;
    switch (Type) {
    case 'N':
        text = (char *)FaxNumber.GetPointer();
        break;
    case 'R':
        recv = vFalse;
        text = "P3ffdffff:ff";
        break;
    case 'Q':
        Type = 'D';
        break;
    case 'B':
    case 'E':
        recv = vTrue;
        break;
    }
    return NotifyFIFO (recv, Type, text);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::WriteLog (tSInt priority, char *text, ...) {
    dhead ("CFaxReceiveDevice::WriteLog", DEBUG_CFaxReceive);
    va_list params;
    va_start (params, text);
    WriteLog (priority, text, params);
    va_end (params);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::vWriteLog (tSInt priority, char *text, va_list args) {
    dhead ("CFaxReceiveDevice::WriteLog", DEBUG_CFaxReceive);
    MainClass->vWriteLog (priority, text, args);
}

/*===========================================================================*\
\*===========================================================================*/

void CFaxReceiveDevice::WriteStatus (char *text) {
    dhead ("CFaxReceiveDevice::WriteStatus", DEBUG_CFaxReceive);
    if (hStatusFile == 0) {
        return;
    }
    flock (fileno (hStatusFile), LOCK_EX);
    rewind (hStatusFile);
    fprintf (hStatusFile, "%s\n", text);
    fflush (hStatusFile);
    ftruncate (fileno (hStatusFile), ftell (hStatusFile));
    flock (fileno (hStatusFile), LOCK_UN);
}

/*===========================================================================*\
\*===========================================================================*/

